/* disp_aggreg.c
 * Disspatcher-aggregator functions protocols
 *
 * $Id: disp_aggreg.c,v 1.4 2007/11/15 19:53:34 costa Exp $
 *
 * Xplico - Internet Traffic Decoder
 * By Gianluca Costa <g.costa@xplico.org>
 * Copyright 2007 Gianluca Costa & Andrea de Franceschi. Web: www.xplico.org
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <stdio.h>

#include "dmemory.h"
#include "log.h"
#include "pei.h"
#include "disp_aggreg.h"
#include "dispatch_dev.h"
#include "proto.h"


static int DispAggrSend(int sock, pei *ppei)
{
    int wr, dim, i;
    pei_component *cmpt;
    pstack_f *stack;
    enum ftype tp;
    ftval attr;

    /* send all data */
    wr = write(sock, ppei, sizeof(pei));
    if (wr == -1)
        return -1;
    cmpt = ppei->components;
    while (cmpt != NULL) {
        wr = write(sock, cmpt, sizeof(pei_component));
        if (cmpt->strbuf != NULL) {
            dim = strlen(cmpt->strbuf) + 1;
            wr = write(sock, &dim, sizeof(dim));
            wr = write(sock, cmpt->strbuf, dim);
        }
        if (cmpt->name != NULL) {
            dim = strlen(cmpt->name) + 1;
            wr = write(sock, &dim, sizeof(dim));
            wr = write(sock, cmpt->name, dim);
        }
        if (cmpt->file_path != NULL) {
            dim = strlen(cmpt->file_path) + 1;
            wr = write(sock, &dim, sizeof(dim));
            wr = write(sock, cmpt->file_path, dim);
        }
        cmpt = cmpt->next;
    }
    stack = ppei->stack;
    while (stack != NULL) {
        dim = ProtFrameSize(stack->pid);
        wr = write(sock, &stack->pid, sizeof(int));
        wr = write(sock, stack, dim);
        i = 0;
        tp = ProtAttrType(stack->pid, i);
        while (tp != FT_NONE) {
            if (tp == FT_STRING) {
                ProtGetAttr(stack, i, &attr);
                if (attr.str != NULL) {
                    dim = strlen(attr.str) + 1;
                    wr = write(sock, &dim, sizeof(dim));
                    wr = write(sock, attr.str, dim);
                }
                else {
                    dim = 0;
                    wr = write(sock, &dim, sizeof(dim));
                }
            }
            i++;
            tp = ProtAttrType(stack->pid, i);
        }
        stack = stack->pfp;
    }

    return 0;
}


static int DispAggConn(aggr_con *con)
{
    struct hostent *host;
    struct sockaddr_in servAddr;
    int rc;

    /* up connections */
    host = gethostbyname(con->host);
    if (host == NULL) {
        printf("Host %s don't exist\n", con->host);
        return -1;
    }
    servAddr.sin_family = host->h_addrtype;
    memcpy((char *) &servAddr.sin_addr.s_addr, host->h_addr_list[0], host->h_length);
    servAddr.sin_port = htons(con->port);
    /* create socket */
    con->sock = socket(AF_INET, SOCK_STREAM, 0);
    if (con->sock < 0) {
        printf("Cannot open socket\n");
        return -1;
    }

    rc = connect(con->sock, (struct sockaddr *) &servAddr, sizeof(servAddr));
    if (rc < 0) {
        printf("Cannot connect to '%s:%u'\n", con->host, con->port);
        close(con->sock);
        con->sock = -1;
        return -1;
    }
    con->wait = FALSE;

    return 0;
}


pei *DispAggrGetPei(int sock)
{
    int rd, dim, pid, i;
    pei *ppei;
    pei_component **cmpt;
    pstack_f **stack;
    enum ftype tp;
    ftval attr;

    ppei = DMemMalloc(sizeof(pei));
    PeiInit(ppei);

    rd = read(sock, ppei, sizeof(pei));
    if (rd == sizeof(pei)) {
        cmpt = &(ppei->components);
        while (*cmpt != NULL) {
            *cmpt = DMemMalloc(sizeof(pei_component));
            rd = read(sock, *cmpt, sizeof(pei_component));
            if (rd == sizeof(pei_component)) {
                if ((*cmpt)->strbuf != NULL) {
                    rd = read(sock, &dim, sizeof(dim));
                    (*cmpt)->strbuf = DMemMalloc(dim);
                    rd = read(sock, (*cmpt)->strbuf, dim);
                }
                if ((*cmpt)->name != NULL) {
                    rd = read(sock, &dim, sizeof(dim));
                    (*cmpt)->name = DMemMalloc(dim);
                    rd = read(sock, (*cmpt)->name, dim);
                }
                if ((*cmpt)->file_path != NULL) {
                    rd = read(sock, &dim, sizeof(dim));
                    (*cmpt)->file_path = DMemMalloc(dim);
                    rd = read(sock, (*cmpt)->file_path, dim);
                }
                cmpt = &((*cmpt)->next);
            }
            else {
                *cmpt = NULL;
            }
        }

        stack = &(ppei->stack);
        while (*stack != NULL) {
            rd = read(sock, &pid, sizeof(int));
            dim = ProtFrameSize(pid);
            *stack = ProtCreateFrame(pid);
            rd = read(sock, (*stack), dim);
            i = 0;
            tp = ProtAttrType((*stack)->pid, i);
            while (tp != FT_NONE) {
                if (tp == FT_STRING) {
                    rd = read(sock, &dim, sizeof(dim));
                    if (dim) {
                        attr.str = DMemMalloc(dim);
                        rd = read(sock, attr.str, dim);
                    }
                    else
                        attr.str = NULL;
                    ProtInsAttr(*stack, i, &attr);
                    DMemFree(attr.str);
                }
                i++;
                tp = ProtAttrType((*stack)->pid, i);
            }
            stack = &((*stack)->pfp);
        }
    }
    else {
        DMemFree(ppei);
        ppei = NULL;
    }

    return ppei;
}


int DispAggrPutPei(pei *ppei)
{
    aggr_con *con;
    pei_list *peil;

    if (ppei != NULL) {
        con = DispatAggr(ppei->prot_id);

        if (con != NULL) {
            if (ppei->ret == TRUE) {
                LogPrintf(LV_WARNING, "Pei with return!");
#ifdef XPL_CHECK_CODE
                PeiPrint(ppei);
                ProtStackFrmDisp(ppei->stack, TRUE);
                exit(-1);
#endif
            }

            LogPrintf(LV_DEBUG, "Pei send to aggregator");
            /* lock */
            pthread_mutex_lock(con->mux);
            /* check if aggregator off */
            if (con->wait == TRUE) {
                /* put pei in queue */
                peil = DMemMalloc(sizeof(pei_list));
                peil->ppei = ppei;
                peil->nxt = NULL;
                ppei = NULL;
                if (con->peilast != NULL) {
                    con->peilast->nxt = peil;
                    con->peilast = peil;
                }
                else {
#ifdef XPL_CHECK_CODE
                    if (con->peil != NULL) {
                        LogPrintf(LV_FATAL, "Aggregator queue error");
                        exit(-1);
                    }
#endif                    
                    con->peil = peil;
                    con->peilast = peil;
                }
                /* reconnect */
                if (DispAggConn(con) == 0) {
                    /* send all pei */
                    peil = con->peil;
                    while (peil != NULL) {
                        if (DispAggrSend(con->sock, peil->ppei) != 0) {
                            DispatAggrOff(peil->ppei->prot_id);
                            peil = NULL;
                        }
                        else {
                            con->peil = peil->nxt;
                            PeiFree(peil->ppei);
                            DMemFree(peil);
                            peil = con->peil;
                            if (peil == NULL) {
                                con->peilast = NULL;
                            }
                        }
                    }
                }
            }
            else {
                /* send to aggregator */
                if (DispAggrSend(con->sock, ppei) != 0) {
                    DispatAggrOff(ppei->prot_id);
                    /* put pei in queue */
                    peil = DMemMalloc(sizeof(pei_list));
                    peil->ppei = ppei;
                    peil->nxt = NULL;
                    ppei = NULL;
#ifdef XPL_CHECK_CODE
                    if (con->peil != NULL) {
                        LogPrintf(LV_FATAL, "Aggregator queue error");
                        exit(-1);
                    }
#endif                    
                    con->peil = peil;
                    con->peilast = peil;
                }
            }
            /* unlock */
            pthread_mutex_unlock(con->mux);
            
            if (ppei != NULL) {
                /* free memory */
                PeiFree(ppei);
            }
            return 0;
        }
    }

    return -1;
}


int DispAggrInit(void)
{
    aggr_con *con;

    /* contact aggregators */
    con = DispatAggrOffLine();
    while (con != NULL) {
        if (DispAggConn(con) != 0) {
            return -1;
        }

        /* next server */
        con = DispatAggrOffLine();
    }

    return 0;
}


int DispAggrEnd(void)
{
    aggr_con *con;
    pei_list *peil;

    /* contact aggregators */
    con = DispatAggrWait();
    while (con != NULL) {
        while (DispAggConn(con) != 0) {
            sleep(1);
        }
        /* send all pei */
        peil = con->peil;
        while (peil != NULL) {
            if (DispAggrSend(con->sock, peil->ppei) != 0) {
                DispatAggrOff(peil->ppei->prot_id);
                peil = NULL;
            }
            else {
                con->peil = peil->nxt;
                PeiFree(peil->ppei);
                DMemFree(peil);
                peil = con->peil;
                if (peil == NULL) {
                    con->peilast = NULL;
                }
            }
        }

        /* next server */
        con = DispatAggrWait();
    }
    return 0;
}
