/* lite.c
 * Xplico System dispatcher module for DEFT Linux distribution
 *
 * $Id: deft.c,v 1.1 2007/09/08 07:26:33 costa Exp $
 *
 * Xplico - Internet Traffic Decoder
 * By Gianluca Costa <g.costa@xplico.org>
 * Copyright 2007 Gianluca Costa & Andrea de Franceschi. Web: www.xplico.org
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <pthread.h>
#include <fcntl.h>
#include <sqlite.h> /* CakePHP is only compatible  with sqlite2 */

#include "proto.h"
#include "log.h"
#include "pei.h"
#include "dmemory.h"
#include "lite.h"
#include "gearth.h"
#include "dnsdb.h"


#define DB_ALWAYS_OPEN    1

/* pol */
static int pol_id;
static int pol_polid_id;
static int pol_sesid_id;
static int pol_filename_id;
/* ip */
static int ip_id;
static int ip_src_id;
static int ip_dst_id;
static int ipv6_id;
static int ipv6_src_id;
static int ipv6_dst_id;
/* pop id */
static int pop_id;
static int pei_pop_user_id;
static int pei_pop_pswd_id;
static int pei_pop_eml_id;
/* smtp id */
static int smtp_id;
static int pei_smtp_to_id;
static int pei_smtp_from_id;
static int pei_smtp_eml_id;
/* imap id */
static int imap_id;
static int pei_imap_user_id;
static int pei_imap_pswd_id;
static int pei_imap_eml_id;
/* http id */
static int http_id;
static int pei_http_url_id;
static int pei_http_client_id;
static int pei_http_host_id;
static int pei_http_content_type_id;
static int pei_http_method_id;
static int pei_http_status_id;
static int pei_http_req_header_id;
static int pei_http_req_body_id;
static int pei_http_res_header_id;
static int pei_http_res_body_id;
/* sip */
static int sip_id;
static int pei_sip_from_id;
static int pei_sip_to_id;
/* ftp */
static int ftp_id;
static int pei_ftp_url_id;
static int pei_ftp_user_id;
static int pei_ftp_pswd_id;
static int pei_ftp_cmd_id;
static int pei_ftp_file_in_id;
static int pei_ftp_file_out_id;
static int pei_ftp_file_offset_id;
static int pei_ftp_down_n_id;
static int pei_ftp_up_n_id;
/* ipp */
static int ipp_id;
static int pei_ipp_url_id;
static int pei_ipp_pdf_id;
static int pei_ipp_pcl_id;
/* pjl */
static int pjl_id;
static int pei_pjl_url_id;
static int pei_pjl_pdf_id;
static int pei_pjl_pcl_id;
/* mms */
static int mms_id;
static int pei_mms_url_id;
static int pei_mms_from_id;
static int pei_mms_to_id;
static int pei_mms_cc_id;
static int pei_mms_bcc_id;
static int pei_mms_part_id;
static int pei_mms_raw_id;
/* */
static int tftp_id;
static int pei_tftp_url_id;
static int pei_tftp_cmd_id;
static int pei_tftp_file_in_id;
static int pei_tftp_file_out_id;
static int pei_tftp_down_n_id;
static int pei_tftp_up_n_id;


/* sqlite */
static sqlite *db;              /* sqlite DB */
static pthread_mutex_t db_mux;  /* mutex to access db */
static pthread_mutex_t feed_mux;  /* mutex to access feed table */
static char query_l[2*XS_QUERY_DIM]; /* to be used inside db_mux mutex lock */

/* geomap */
static unsigned long geo_id; /* geo session number, in this case we have only one session */

/* host */
static host_id * volatile host;
static volatile unsigned long host_num;
static volatile unsigned long host_dim;
static pthread_mutex_t host_mux;


static int DispQuery(char *query, int *id);

static int DispHostExt(void)
{
    host_id *new;

    /* the mutex is in already locked */
    new = xrealloc(host, sizeof(host_id)*(host_dim + XS_HOST_ID_ADD));
    if (new == NULL)
        return -1;
    memset(new+sizeof(host_id)*(host_dim), 0, sizeof(host_id)*XS_HOST_ID_ADD);
    host = new;
    host_dim += XS_HOST_ID_ADD;
    
    return 0;
}


static int DispHostSrch(ftval *ip, enum ftype type)
{
    int i, ret;

    pthread_mutex_lock(&host_mux);
    for (i=0; i != host_num; i++) {
        if (host[i].type == type) {
            if (FTCmp(ip, &(host[i].ip), type, FT_OP_EQ) == 0) {
                ret = host[i].id;
                pthread_mutex_unlock(&host_mux);
                return ret;
            }
        }
    }
    /* mutex unlock at DispHostIns */

    return -1;
}


static int DispHostIns(ftval *ip, enum ftype type, int db_id)
{
    if (host_num == host_dim) {
        if (DispHostExt() != 0) {
            pthread_mutex_unlock(&host_mux);
            return -1;
        }
        host[host_num].id = db_id;
        FTCopy(&(host[host_num].ip), ip, type);
        host[host_num].type = type;
        host_num++;
    }
    pthread_mutex_unlock(&host_mux);

    return 0;
}


static int DispHostDb(const char *ip, int pol, int sol)
{
    int ret, res;
    int ncolumns; /* OUT: Number of columns in result */
    const char **pazValue;  /* OUT: Column data */
    const char **pazColName; /* OUT: Column names and datatypes */
    sqlite_vm *pVm;
    char *err;
    
    ret = -1;
    err = NULL;
    pthread_mutex_lock(&db_mux);
#if DB_ALWAYS_OPEN == 0
    while ((db = sqlite_open(XS_DB_FILE, 0, &err)) == NULL)
        sched_yield();
#endif
    sprintf(query_l, XS_QUERY_SOURCE_SEARCH, sol, ip);
    res = sqlite_compile(db, query_l, NULL, &pVm, &err);
    if (res == SQLITE_OK) {
        res = sqlite_step(pVm, &ncolumns, &pazValue, &pazColName);
        if (res != SQLITE_ROW)
            sqlite_finalize(pVm, &err);
    }
    if (err != NULL) {
        free(err);
        err = NULL;
    }
    if (res != SQLITE_ROW) {
#if DB_ALWAYS_OPEN == 0
        sqlite_close(db);
#endif
        pthread_mutex_unlock(&db_mux);
    }
    else {
        ret = atoi(pazValue[0]);
        sqlite_finalize(pVm, &err);
#if DB_ALWAYS_OPEN == 0
        sqlite_close(db);
#endif
        pthread_mutex_unlock(&db_mux);
    }
    if (err != NULL) {
        free(err);
        err = NULL;
    }

    return ret;
}


static int DispHostDbIns(const char *ip, char *name, int pol, int sol)
{
    int id;
    char query[XS_QUERY_DIM];

    sprintf(query, XS_QUERY_SOURCE_TEMPLATE, sol, pol, ip, name);
    if (DispQuery(query, &id) != 0) {
        printf("query: %s\n", query);
    }

    return id;
}


static void DispFlowInfo(char *fname, const pstack_f *stack)
{
    char *xmlog;
    int fd;

    if (stack != NULL) {
        /* xml flow */
        xmlog = ProtStackFrmXML(stack);
        
        /* save_xml_file */
        fd = open(fname, O_CREAT|O_RDWR, 0x01B6);
        if (fd != -1) {
            write(fd, xmlog, strlen(xmlog));
            close(fd);
        }
        xfree(xmlog);
    }
}


static char* DispUrlNorm(char *url)
{
    char *app, a, b;
    int i;

    app = strchr(url, '\'');
    while (app != NULL) {
        i = 0;
        a = '\\';
        while (a != '\0') {
            b = app[i+1];
            app[i] = a;
            a = b;
            i++;
        }
        app[i] = a;

        app += 2; /* possible ' */
        app = strchr(url, '\'');
    }
    
    return url;
}


static int DispCopy(char *old, char *new, bool rm)
{
    bool cp;
    char *buff;
    FILE *in, *out;
    size_t len;

    cp = !rm;
    if (rm) {
        /* rename */
        if (rename(old, new) != 0)
            cp = TRUE;
        else
            rm = FALSE;
    }
    if (cp) {
        /* copy */        
        in = fopen(old, "r");
        if (in != NULL) {
            out = fopen(new, "w");
            if (out != NULL) {
                buff = xmalloc(XS_CP_BUFFER);
                if (buff != NULL) {
                    while ((len = fread(buff, XS_CP_BUFFER, 1, in)) != 0)
                        fwrite(buff, len, 1, out);
                    xfree(buff);
                }
                fclose(out);
            }
            else {
                LogPrintf(LV_WARNING, "Unable to open file (%s)", new);
            }
            fclose(in);
        }
        else {
            if (rm)
                LogPrintf(LV_WARNING, "Unable to open file (%s)", old);
            else {
                /* create file */
                out = fopen(new, "w");
                if (out != NULL) {
                    fclose(out);
                }
            }
        }
    }
    if (rm) {
        /* remove */
        remove(old);
    }

    return 0;
}


static int DispQuery(char *query, int *id)
{
    int ret;
    char *err = NULL;
    sqlite *dbl = NULL;

    pthread_mutex_lock(&db_mux);

    err = NULL;
#if DB_ALWAYS_OPEN
    dbl = db;
#else
    while ((dbl = sqlite_open(XS_DB_FILE, 0, &err)) == NULL) {
        pthread_mutex_unlock(&db_mux);
        //sched_yield();
        pthread_mutex_lock(&db_mux);
    }
    if (err != NULL) {
        LogPrintf(LV_ERROR, "DB error: %s", err);
        free(err);
        err = NULL;
    }
    sqlite_exec(dbl, "PRAGMA synchronous=FULL", NULL, 0, 0);
    if (sqlite_exec(dbl, "PRAGMA integrity_check", NULL, 0, &err) != 0) {
        if (err != NULL) {
            LogPrintf(LV_ERROR, "DB error: %s", err);
            free(err);
            err = NULL;
        }
    }
    err = NULL;
#endif

    while (sqlite_exec(dbl, query, 0, 0, &err) == SQLITE_BUSY) {
        if (err != NULL)
            free(err);
        err = NULL;
    }
    if (err != NULL) {
        LogPrintf(LV_ERROR, "Query: %s", query);
        LogPrintf(LV_DEBUG, "Error: %s", err);
        free(err);
    }
    if (id != NULL) {
        /* return record id */
        *id = sqlite_last_insert_rowid(dbl);
    }

#if DB_ALWAYS_OPEN == 0
    sqlite_close(dbl);
#endif

    pthread_mutex_unlock(&db_mux);

    ret = 0;

    return ret;
}


static int DispStrCnv(char *dst, char *src, int len)
{
    int i, n, dim;

    i = 0;
    n = 0;
    dim = strlen(dst);
    if (dim != 0) {
        n = dim ;
        dst[n++] = ',';
        dst[n++] = ' ';
    }
    len = len - dim;

    while (i<len && src[i] != '\0') {
        if (src[i] == '\'') {
            dst[n] = '\'';
            n++;
        }
        if (src[i] != '\r' && src[i] != '\n') {
            dst[n] = src[i];
            n++;
        }
        i++;
    }

    return 0;
}


static char *DispLabelCnv(char *src, char *dst)
{
    int i, n, len;

    i = 0;
    n = 0;
    len = strlen(src);
    while (i<len && src[i] != '\0') {
        if (src[i] == '\'') {
            dst[n] = '\'';
            n++;
        }
        dst[n] = src[i];
        n++;
        i++;
    }
    dst[n] = '\0';

    return dst;
}


static int DispMimeParse(char *file, char *subject, char *from, char *to)
{
    char cmd[XS_QUERY_DIM];
    char dump[XS_STR_DIM];
    char none[XS_STR_NONE];
    FILE *fp;
    int data, ret;
    
    /* mime dump */
    sprintf(dump, "%s/dump_%i_%p.txt", ProtTmpDir(), getpid(), file);
    sprintf(cmd, "%s %s > %s", XS_MIME_PARSER, file, dump);
    system(cmd);

    /* read file */
    fp = fopen(dump, "r");
    if (fp) {
        ret = 0;
        data = 0;
        while (fgets(cmd, XS_QUERY_DIM, fp) != NULL) {
            if (strncmp(cmd, XS_MIME_TO, 3) == 0) {
                data = 1;
            }
            else if (strncmp(cmd, XS_MIME_FROM, 5) == 0) {
                data = 2;
            }
            else if (strncmp(cmd, XS_MIME_SUBJECT, 8) == 0) {
                data = 3;
            }
            else {
                switch (data) {
                case 1:
                    if (to != NULL) {
                        DispStrCnv(to, cmd, XS_STR_DIM);
                        if (strcmp(to, "None") == 0)
                            ret++;
                    }
                    else {
                        DispStrCnv(none, cmd, XS_STR_NONE);
                        if (strcmp(none, "None") == 0)
                            ret++;
                    }
                    break;
                    
                case 2:
                    if (from != NULL) {
                        DispStrCnv(from, cmd, XS_STR_DIM);
                        if (strcmp(from, "None") == 0)
                            ret++;
                    }
                    else {
                        DispStrCnv(none, cmd, XS_STR_NONE);
                        if (strcmp(none, "None") == 0)
                            ret++;
                    }
                    break;
                    
                case 3:
                    DispStrCnv(subject, cmd, XS_STR_DIM);
                    if (strcmp(subject, "None") == 0)
                        ret++;
                    break;
                    
                default:
                    break;
                }
            }
            memset(cmd, 0, XS_QUERY_DIM);
            if (ret == 3)
                ret = -1;
        }
        fclose(fp);
    }
    else {
        ret = -1;
    }
    remove(dump);

    return ret;
}


static int DispPop(pei *ppei)
{
    pei_component *cmpn;
    char query[XS_QUERY_DIM];
    char rep[XS_QUERY_DIM];
    char subject[XS_STR_DIM];
    char from[XS_STR_DIM];
    char to[XS_STR_DIM];
    char flow_info[XS_STR_PATH];
    int pol, sess, src_id;
    char *user;
    char *pwd;
    char *path;
    struct stat info;
    const pstack_f *frame;
    ftval val, ip;
    char *name;

    /* search pol and session */
    frame = ProtStackSearchProt(ppei->stack, pol_id);
    if (frame) {
        ProtGetAttr(frame, pol_polid_id, &val);
        pol = val.int32;
        ProtGetAttr(frame, pol_sesid_id, &val);
        sess = val.int32;
    }
    else {
        sess = pol = 1;
    }
    /* search source ip */
    src_id = -1;
    frame = ProtStackSearchProt(ppei->stack, ip_id);
    if (frame) {
        ProtGetAttr(frame, ip_src_id, &ip);
        src_id = DispHostSrch(&ip, FT_IPv4);
        if (src_id == -1) {
            /* search in db */
            FTString(&ip, FT_IPv4, flow_info);
            src_id = DispHostDb(flow_info, pol, sess);
            if (src_id == -1) {
                /* insert record */
                DnsDbSeach(&ip, FT_IPv4, query, XS_QUERY_DIM);
                src_id = DispHostDbIns(flow_info, query, pol, sess);
            }
            DispHostIns(&ip, FT_IPv4, src_id);
        }
    }
    else if (ipv6_id != -1) {
        frame = ProtStackSearchProt(ppei->stack, ipv6_id);
        if (frame) {
            ProtGetAttr(frame, ipv6_src_id, &ip);
            src_id = DispHostSrch(&ip, FT_IPv4);
            if (src_id == -1) {
                /* search in db */
                FTString(&ip, FT_IPv6, flow_info);
                src_id = DispHostDb(flow_info, pol, sess);
                if (src_id == -1) {
                    /* insert record */
                    DnsDbSeach(&ip, FT_IPv6, query, XS_QUERY_DIM);
                    src_id = DispHostDbIns(flow_info, query, pol, sess);
                }
                DispHostIns(&ip, FT_IPv6, src_id);
            }
        }
    }
    path = pwd = user = NULL;
    cmpn = ppei->components;
    while (cmpn != NULL) {
        if (cmpn->eid == pei_pop_user_id) {
            user = cmpn->strbuf;
        }
        if (cmpn->eid == pei_pop_pswd_id) {
            pwd = cmpn->strbuf;

        }
        if (cmpn->eid == pei_pop_eml_id) {
            path = cmpn->file_path;
        }
        cmpn = cmpn->next;
    }

    /* compose query and insert record */
    if (path && stat(path, &info) == 0) {
        /* new path */
        name = strrchr(path, '/');
        name++;
        sprintf(rep, XS_MAIL_DIR_PATH"/%s", pol, sess, name);
        rename(path, rep);
        /* flow info */
        sprintf(flow_info, XS_MAIL_DIR_PATH"/flow_%s.xml", pol, sess, name);
        DispFlowInfo(flow_info, ppei->stack);
        /* parse mime */
        memset(subject, 0, XS_STR_DIM);
        memset(from, 0, XS_STR_DIM);
        memset(to, 0, XS_STR_DIM);
        if (DispMimeParse(rep, subject, from, to) == 0) {
            /* query */
            sprintf(query, XS_QUERY_POP_TEMPLATE, sess, pol, src_id, time(NULL)-ppei->time_cap, info.st_size, flow_info,
                    user, pwd, from, to, subject, rep);
            if (DispQuery(query, NULL) != 0) {
                printf("query: %s\n", query);
            }
        }
    }

    return 0;
}


static int DispSmtp(pei *ppei)
{
    pei_component *cmpn;
    char query[XS_QUERY_DIM];
    char rep[XS_QUERY_DIM];
    char subject[XS_STR_DIM];
    char *from;
    char to[XS_STR_DIM];
    char flow_info[XS_STR_PATH];
    int pol;
    int sess, src_id;
    char *path;
    struct stat info;
    const pstack_f *frame;
    ftval val, ip;
    char *name;

    /* search pol and session */
    frame = ProtStackSearchProt(ppei->stack, pol_id);
    if (frame) {
        ProtGetAttr(frame, pol_polid_id, &val);
        pol = val.int32;
        ProtGetAttr(frame, pol_sesid_id, &val);
        sess = val.int32;
    }
    else {
        sess = pol = 1;
    }
    /* search source ip */
    src_id = -1;
    frame = ProtStackSearchProt(ppei->stack, ip_id);
    if (frame) {
        ProtGetAttr(frame, ip_src_id, &ip);
        src_id = DispHostSrch(&ip, FT_IPv4);
        if (src_id == -1) {
            /* search in db */
            FTString(&ip, FT_IPv4, flow_info);
            src_id = DispHostDb(flow_info, pol, sess);
            if (src_id == -1) {
                /* insert record */
                DnsDbSeach(&ip, FT_IPv4, query, XS_QUERY_DIM);
                src_id = DispHostDbIns(flow_info, query, pol, sess);
            }
            DispHostIns(&ip, FT_IPv4, src_id);
        }
    }
    else if (ipv6_id != -1) {
        frame = ProtStackSearchProt(ppei->stack, ipv6_id);
        if (frame) {
            ProtGetAttr(frame, ipv6_src_id, &ip);
            src_id = DispHostSrch(&ip, FT_IPv4);
            if (src_id == -1) {
                /* search in db */
                FTString(&ip, FT_IPv6, flow_info);
                src_id = DispHostDb(flow_info, pol, sess);
                if (src_id == -1) {
                    /* insert record */
                    DnsDbSeach(&ip, FT_IPv6, query, XS_QUERY_DIM);
                    src_id = DispHostDbIns(flow_info, query, pol, sess);
                }
                DispHostIns(&ip, FT_IPv6, src_id);
            }
        }
    }
    path = NULL;
    to[0] = '\0';
    cmpn = ppei->components;
    while (cmpn != NULL) {
        if (cmpn->eid == pei_smtp_to_id) {
            if (to[0] != '\0') {
                strcat(to, ", ");
            }
            strcat(to, cmpn->strbuf);
        }
        if (cmpn->eid == pei_smtp_from_id) {
            from = cmpn->strbuf;
        }
        if (cmpn->eid == pei_smtp_eml_id) {
            path = cmpn->file_path;
        }
        cmpn = cmpn->next;
    }

    /* compose query and insert record */
    if (path &&stat(path, &info) == 0) {
        /* new path */
        name = strrchr(path, '/');
        name++;
        sprintf(rep, XS_MAIL_DIR_PATH"/%s", pol, sess, name);
        rename(path, rep);
        /* flow info */
        sprintf(flow_info, XS_MAIL_DIR_PATH"/flow_%s.xml", pol, sess, name);
        DispFlowInfo(flow_info, ppei->stack);
        /* parse mime */
        memset(subject, 0, XS_STR_DIM);
        if (DispMimeParse(rep, subject, NULL, NULL) == 0) {
            /* query */
            sprintf(query, XS_QUERY_SMTP_TEMPLATE, sess, pol, src_id, time(NULL)-ppei->time_cap, info.st_size, flow_info,
                    from, to, subject, rep);
            if (DispQuery(query, NULL) != 0) {
                printf("query: %s\n", query);
            }
        }
    }

    return 0;
}


static int DispImap(pei *ppei)
{
    pei_component *cmpn;
    char query[XS_QUERY_DIM];
    char rep[XS_QUERY_DIM];
    char subject[XS_STR_DIM];
    char from[XS_STR_DIM];
    char to[XS_STR_DIM];
    char flow_info[XS_STR_PATH];
    int pol, sess, src_id;
    char *user;
    char *pwd;
    char *path;
    struct stat info;
    const pstack_f *frame;
    ftval val, ip;
    char *name;

    /* search pol and session */
    frame = ProtStackSearchProt(ppei->stack, pol_id);
    if (frame) {
        ProtGetAttr(frame, pol_polid_id, &val);
        pol = val.int32;
        ProtGetAttr(frame, pol_sesid_id, &val);
        sess = val.int32;
    }
    else {
        sess = pol = 1;
    }
    /* search source ip */
    src_id = -1;
    frame = ProtStackSearchProt(ppei->stack, ip_id);
    if (frame) {
        ProtGetAttr(frame, ip_src_id, &ip);
        src_id = DispHostSrch(&ip, FT_IPv4);
        if (src_id == -1) {
            /* search in db */
            FTString(&ip, FT_IPv4, flow_info);
            src_id = DispHostDb(flow_info, pol, sess);
            if (src_id == -1) {
                /* insert record */
                DnsDbSeach(&ip, FT_IPv4, query, XS_QUERY_DIM);
                src_id = DispHostDbIns(flow_info, query, pol, sess);
            }
            DispHostIns(&ip, FT_IPv4, src_id);
        }
    }
    else if (ipv6_id != -1) {
        frame = ProtStackSearchProt(ppei->stack, ipv6_id);
        if (frame) {
            ProtGetAttr(frame, ipv6_src_id, &ip);
            src_id = DispHostSrch(&ip, FT_IPv4);
            if (src_id == -1) {
                /* search in db */
                FTString(&ip, FT_IPv6, flow_info);
                src_id = DispHostDb(flow_info, pol, sess);
                if (src_id == -1) {
                    /* insert record */
                    DnsDbSeach(&ip, FT_IPv6, query, XS_QUERY_DIM);
                    src_id = DispHostDbIns(flow_info, query, pol, sess);
                }
                DispHostIns(&ip, FT_IPv6, src_id);
            }
        }
    }
    path = pwd = user = NULL;
    cmpn = ppei->components;
    while (cmpn != NULL) {
        if (cmpn->eid == pei_imap_user_id) {
            user = cmpn->strbuf;
        }
        if (cmpn->eid == pei_imap_pswd_id) {
            pwd = cmpn->strbuf;

        }
        if (cmpn->eid == pei_imap_eml_id) {
            path = cmpn->file_path;
        }
        cmpn = cmpn->next;
    }

    /* compose query and insert record */
    if (path && stat(path, &info) == 0) {
        /* new path */
        name = strrchr(path, '/');
        name++;
        sprintf(rep, XS_MAIL_DIR_PATH"/%s", pol, sess, name);
        rename(path, rep);
        /* flow info */
        sprintf(flow_info, XS_MAIL_DIR_PATH"/flow_%s.xml", pol, sess, name);
        DispFlowInfo(flow_info, ppei->stack);
        /* parse mime */
        memset(subject, 0, XS_STR_DIM);
        memset(from, 0, XS_STR_DIM);
        memset(to, 0, XS_STR_DIM);
        if (DispMimeParse(rep, subject, from, to) == 0) {
            /* query */
            sprintf(query, XS_QUERY_IMAP_TEMPLATE, sess, pol, src_id, time(NULL)-ppei->time_cap, info.st_size, flow_info,
                    user, pwd, from, to, subject, rep);
            if (DispQuery(query, NULL) != 0) {
                printf("query: %s\n", query);
            }
        }
    }

    return 0;
}


static int DispHttp(pei *ppei)
{
    pei_component *cmpn;
    char query[2*XS_QUERY_DIM];
    char flow_info[XS_STR_PATH];
    char url[2*XS_QUERY_DIM];
    char rq_header[XS_STR_PATH], rq_body[XS_STR_PATH],
        rs_header[XS_STR_PATH], rs_body[XS_STR_PATH], tmp[XS_STR_PATH];
    char buffer[XS_QUERY_DIM];
    int pol, sess, dim, ret, id, res, src_id;
    char *host, *host_t, *urlp, *path;
    char *cont_type, *title, *link, *end, *err;
    sqlite_vm *pVm;
    int ncolumns; /* OUT: Number of columns in result */
    const char **pazValue;  /* OUT: Column data */
    const char **pazColName; /* OUT: Column names and datatypes */
    char *rep, *method, *response;
    int rq_bd_size, rs_bd_size;
    const pstack_f *frame;
    ftval val, ip;
    char *name;
    bool http;
    FILE *fp;

    /* search pol and session */
    frame = ProtStackSearchProt(ppei->stack, pol_id);
    if (frame) {
        ProtGetAttr(frame, pol_polid_id, &val);
        pol = val.int32;
        ProtGetAttr(frame, pol_sesid_id, &val);
        sess = val.int32;
    }
    else {
        sess = pol = 1;
    }
    /* search source ip */
    src_id = -1;
    frame = ProtStackSearchProt(ppei->stack, ip_id);
    if (frame) {
        ProtGetAttr(frame, ip_src_id, &ip);
        src_id = DispHostSrch(&ip, FT_IPv4);
        if (src_id == -1) {
            /* search in db */
            FTString(&ip, FT_IPv4, buffer);
            src_id = DispHostDb(buffer, pol, sess);
            if (src_id == -1) {
                /* insert record */
                DnsDbSeach(&ip, FT_IPv4, url, XS_QUERY_DIM);
                src_id = DispHostDbIns(buffer, url, pol, sess);
            }
            DispHostIns(&ip, FT_IPv4, src_id);
        }
    }
    else if (ipv6_id != -1) {
        frame = ProtStackSearchProt(ppei->stack, ipv6_id);
        if (frame) {
            ProtGetAttr(frame, ipv6_src_id, &ip);
            src_id = DispHostSrch(&ip, FT_IPv4);
            if (src_id == -1) {
                /* search in db */
                FTString(&ip, FT_IPv6, buffer);
                src_id = DispHostDb(buffer, pol, sess);
                if (src_id == -1) {
                    /* insert record */
                    DnsDbSeach(&ip, FT_IPv6, url, XS_QUERY_DIM);
                    src_id = DispHostDbIns(buffer, url, pol, sess);
                }
                DispHostIns(&ip, FT_IPv6, src_id);
            }
        }
    }
    host = host_t = method = response = rep = cont_type = urlp = NULL;
    flow_info[0] = '\0';
    url[0] = '\0';
    rq_bd_size = rs_bd_size = 0;
    memcpy(rq_header, "NULL", 5);
    memcpy(rq_body, "NULL", 5);
    memcpy(rs_header, "NULL", 5);
    memcpy(rs_body, "NULL", 5);
    cmpn = ppei->components;
    http = FALSE;
    err = NULL;
    pVm = NULL;
    pazValue = NULL;
    pazColName = NULL;
    while (cmpn != NULL) {
        if (cmpn->eid == pei_http_url_id) {
            if (strlen(cmpn->strbuf) >= XS_HTTP_URL_LEN && strncmp(cmpn->strbuf, XS_HTTP_URL, XS_HTTP_URL_LEN) == 0) {
                http = TRUE;
            }
            if (host_t != NULL && http == FALSE) {
                sprintf(url, "%s%s", host_t, cmpn->strbuf);
            }
            else {
                if (http == TRUE) {
                    sprintf(url, "%s", cmpn->strbuf+XS_HTTP_URL_LEN);
                }
                else {
                    urlp = cmpn->strbuf;
                }
            }
        }
        else if (cmpn->eid == pei_http_host_id) {
            host = cmpn->strbuf;
            if (urlp != NULL) {
                sprintf(url, "%s%s", cmpn->strbuf, urlp);
            }
            else {
                host_t = cmpn->strbuf;
            }
        }
        else if (cmpn->eid == pei_http_content_type_id) {
            cont_type = cmpn->strbuf;
        }
        else if (cmpn->eid == pei_http_method_id) {
            method = cmpn->strbuf;
        }
        else if (cmpn->eid == pei_http_status_id) {
            response = cmpn->strbuf;
        }
        else if (cmpn->eid == pei_http_req_header_id) {
            /* new path */
            rep = rq_header;
            path = cmpn->file_path;
            name = strrchr(path, '/');
            name++;
            sprintf(rep, XS_HTTP_DIR_PATH"/%s", pol, sess, name);
            rename(path, rep);
        }
        else if (cmpn->eid == pei_http_req_body_id) {
            /* new path */
            rep = rq_body;
            path = cmpn->file_path;
            name = strrchr(path, '/');
            name++;
            sprintf(rep, XS_HTTP_DIR_PATH"/%s", pol, sess, name);
            rename(path, rep);
            rq_bd_size = cmpn->file_size;
        }
        else if (cmpn->eid == pei_http_res_header_id) {
            /* new path */
            rep = rs_header;
            path = cmpn->file_path;
            name = strrchr(path, '/');
            name++;
            sprintf(rep, XS_HTTP_DIR_PATH"/%s", pol, sess, name);
            rename(path, rep);
        }
        else if (cmpn->eid == pei_http_res_body_id) {
            /* new path */
            rep = rs_body;
            path = cmpn->file_path;
            name = strrchr(path, '/');
            name++;
            sprintf(rep, XS_HTTP_DIR_PATH"/%s", pol, sess, name);
            rename(path, rep);
            rs_bd_size = cmpn->file_size;
        }

        cmpn = cmpn->next;
    }

    /* compose query and insert record */
    if (rs_header != NULL) {
        /* normalise url */
        if (strchr(url, '\'') != NULL) {
            DispUrlNorm(url);
        }

        /* flow info */
        sprintf(flow_info, XS_HTTP_DIR_PATH"/flow_%s.xml", pol, sess, name);
        DispFlowInfo(flow_info, ppei->stack);
        /* query */
        sprintf(query, XS_QUERY_WEB_TEMPLATE, sess, pol, src_id, time(NULL)-ppei->time_cap, flow_info,
                url, XS_URL_REL_6, method, response, host, cont_type, rq_header, rq_body, rq_bd_size, rs_header, rs_body, rs_bd_size);
        if (DispQuery(query, NULL) != 0) {
            printf("query: %s\n", query);
        }
    }

    /* if feed */
#if 1
    if (cont_type != NULL && rs_bd_size != 0 && strstr(cont_type, "xml") != NULL) {
        /* read body file */
        sprintf(tmp, "gunzip %s -c > %s/%p.dec 2> /dev/null", rs_body, ProtTmpDir(), rs_body);
        ret = system(tmp);
        sprintf(tmp, "%s/%p.dec", ProtTmpDir(), rs_body);
        if (ret == 0) {
            fp = fopen(tmp, "r");
        }
        else {
            fp = fopen(rs_body, "r");
        }
        if (fp != NULL) {
            dim = fread(buffer, 1, XS_QUERY_DIM, fp);
            buffer[dim-1] = '\0';
            /* RSS */
            if (strstr(buffer, "<rss") != NULL) {
                /* find title */
                title = strstr(buffer, "<title>");
                if (title != NULL) {
                    title += 7; 
                    end = strstr(title, "</");
                    if (end != NULL) {
                        end[0] = '\0';
                        char dst[strlen(title)*2];

                        /* find if already insert */
                        pthread_mutex_lock(&feed_mux); /* this garantee the use of only one entry */
                        pthread_mutex_lock(&db_mux);
#if DB_ALWAYS_OPEN == 0
                        while ((db = sqlite_open(XS_DB_FILE, 0, &err)) == NULL)
                            sched_yield();
#endif
                        sprintf(query, XS_QUERY_FEEDS_SEARCH, sess, DispLabelCnv(title, dst));
                        res = sqlite_compile(db, query, NULL, &pVm, &err);
                        if (res == SQLITE_OK) {
                            res = sqlite_step(pVm, &ncolumns, &pazValue, &pazColName);
                            if (res != SQLITE_ROW)
                                sqlite_finalize(pVm, &err);
                        }
                        if (err != NULL) {
                            free(err);
                            err = NULL;
                        }
                        if (res != SQLITE_ROW) {
#if DB_ALWAYS_OPEN == 0
                            sqlite_close(db);
#endif
                            pthread_mutex_unlock(&db_mux);
                            
                            /* find link */
                            link = strstr(end+1, "<link>");
                            if (link != NULL) {
                                link += 6;
                                end = strstr(link, "</");
                                if (end != NULL) {
                                    end[0] = '\0';
                                    /*printf("Title: %s link: %s\n", title, link);*/
                                }
                            }
                            else {
                                link = "";
                            }
                            /* insert feed */
                            /* query */
                            sprintf(query, XS_QUERY_FEEDS_TEMPLATE, sess, pol, src_id, DispLabelCnv(title, dst), link);
                            if (DispQuery(query, &id) != 0) {
                                printf("query: %s\n", query);
                            }
                            sprintf(query, XS_QUERY_FEEDS_XML_TEMPLATE, sess, pol, src_id, id, time(NULL)-ppei->time_cap, flow_info, url, rs_header, rs_body, rs_bd_size);
                            if (DispQuery(query, NULL) != 0) {
                                printf("query: %s\n", query);
                            }
                        }
                        else {
                            id = atoi(pazValue[0]);
                            sqlite_finalize(pVm, &err);
#if DB_ALWAYS_OPEN == 0
                            sqlite_close(db);
#endif
                            pthread_mutex_unlock(&db_mux);
                            /* insert feed */
                            /* query */
                            sprintf(query, XS_QUERY_FEEDS_XML_TEMPLATE, sess, pol, src_id, id, time(NULL)-ppei->time_cap, flow_info, url, rs_header, rs_body, rs_bd_size);
                            if (DispQuery(query, NULL) != 0) {
                                printf("query: %s\n", query);
                            }
                        }
                        pthread_mutex_unlock(&feed_mux);
                        if (err != NULL) {
                            free(err);
                            err = NULL;
                        }
                    }
                }
            }
            else {
                /* Atom */
                if (strstr(buffer, "<feed") != NULL && strstr(buffer, "http://www.w3.org/2005/Atom") != NULL) {
                    /* find title */
                    title = strstr(buffer, "<title");
                    if (title != NULL) {
                        title = strstr(title, ">");
                    }
                    if (title != NULL) {
                        title += 1;
                        end = strstr(title, "</");
                        if (end != NULL) {
                            end[0] = '\0';
                            char dst[strlen(title)*2];
                            
                            /* find if already insert */
                            pthread_mutex_lock(&feed_mux); /* this garantee the use of only one entry */
                            pthread_mutex_lock(&db_mux);
#if DB_ALWAYS_OPEN == 0
                            while ((db = sqlite_open(XS_DB_FILE, 0, &err)) == NULL)
                                sched_yield();
#endif
                            sprintf(query, XS_QUERY_FEEDS_SEARCH, sess, DispLabelCnv(title, dst));
                            res = sqlite_compile(db, query, NULL, &pVm, &err);
                            if (res == SQLITE_OK) {
                                res = sqlite_step(pVm, &ncolumns, &pazValue, &pazColName);
                                if (res != SQLITE_ROW)
                                    sqlite_finalize(pVm, &err);
                            }
                            if (res != SQLITE_ROW) {
#if DB_ALWAYS_OPEN == 0
                                sqlite_close(db);
#endif
                                pthread_mutex_unlock(&db_mux);
                                
                                /* find link */
                                link = host;
                                /* insert feed */
                                /* query */
                                sprintf(query, XS_QUERY_FEEDS_TEMPLATE, sess, pol, src_id, DispLabelCnv(title, dst), link);
                                if (DispQuery(query, &id) != 0) {
                                    printf("query: %s\n", query);
                                }
                                sprintf(query, XS_QUERY_FEEDS_XML_TEMPLATE, sess, pol, src_id, id, time(NULL)-ppei->time_cap, flow_info, url, rs_header, rs_body, rs_bd_size);
                                if (DispQuery(query, NULL) != 0) {
                                    printf("query: %s\n", query);
                                }
                            }
                            else {
                                id = atoi(pazValue[0]);
                                sqlite_finalize(pVm, &err);
#if DB_ALWAYS_OPEN == 0
                                sqlite_close(db);
#endif
                                pthread_mutex_unlock(&db_mux);
                                /* insert feed */
                                /* query */
                                sprintf(query, XS_QUERY_FEEDS_XML_TEMPLATE, sess, pol, src_id, id, time(NULL)-ppei->time_cap, flow_info, url, rs_header, rs_body, rs_bd_size);
                                if (DispQuery(query, NULL) != 0) {
                                    printf("query: %s\n", query);
                                }
                            }
                            pthread_mutex_unlock(&feed_mux);
                        }
                    }
                }
            }
            fclose(fp);
        }
        remove(tmp);
    }
#endif

    return 0;
}


static int DispSip(pei *ppei)
{
    static int inc = 0;
    pei_component *cmpn;
    char query[XS_QUERY_DIM];
    char *from, *to;
    char flow_info[XS_STR_PATH];
    int pol, sess, src_id;
    const pstack_f *frame;
    ftval val, ip;

    /* search pol and session */
    frame = ProtStackSearchProt(ppei->stack, pol_id);
    if (frame) {
        ProtGetAttr(frame, pol_polid_id, &val);
        pol = val.int32;
        ProtGetAttr(frame, pol_sesid_id, &val);
        sess = val.int32;
    }
    else {
        sess = pol = 1;
    }
    /* search source ip */
    src_id = -1;
    frame = ProtStackSearchProt(ppei->stack, ip_id);
    if (frame) {
        ProtGetAttr(frame, ip_src_id, &ip);
        src_id = DispHostSrch(&ip, FT_IPv4);
        if (src_id == -1) {
            /* search in db */
            FTString(&ip, FT_IPv4, flow_info);
            src_id = DispHostDb(flow_info, pol, sess);
            if (src_id == -1) {
                /* insert record */
                DnsDbSeach(&ip, FT_IPv4, query, XS_QUERY_DIM);
                src_id = DispHostDbIns(flow_info, query, pol, sess);
            }
            DispHostIns(&ip, FT_IPv4, src_id);
        }
    }
    else if (ipv6_id != -1) {
        frame = ProtStackSearchProt(ppei->stack, ipv6_id);
        if (frame) {
            ProtGetAttr(frame, ipv6_src_id, &ip);
            src_id = DispHostSrch(&ip, FT_IPv4);
            if (src_id == -1) {
                /* search in db */
                FTString(&ip, FT_IPv6, flow_info);
                src_id = DispHostDb(flow_info, pol, sess);
                if (src_id == -1) {
                    /* insert record */
                    DnsDbSeach(&ip, FT_IPv6, query, XS_QUERY_DIM);
                    src_id = DispHostDbIns(flow_info, query, pol, sess);
                }
                DispHostIns(&ip, FT_IPv6, src_id);
            }
        }
    }
    cmpn = ppei->components;
    while (cmpn != NULL) {
        if (cmpn->eid == pei_sip_to_id) {
            to = cmpn->strbuf;
        }
        else if (cmpn->eid == pei_sip_from_id) {
            from = cmpn->strbuf;
        }
        cmpn = cmpn->next;
    }

    /* flow info */
    sprintf(flow_info, XS_SIP_DIR_PATH"/flow_%lu_%i.xml", pol, sess, time(NULL), inc++);
    DispFlowInfo(flow_info, ppei->stack);

    /* compose query and insert record */
    sprintf(query, XS_QUERY_SIP_TEMPLATE, sess, pol, src_id, time(NULL)-ppei->time_cap, flow_info,
            from, to);
    if (DispQuery(query, NULL) != 0) {
        printf("query: %s\n", query);
    }

    return 0;
}


static int DispFtp(pei *ppei)
{
    pei_component *cmpn;
    char query[XS_QUERY_DIM];
    char flow_info[XS_STR_PATH];
    char rep[XS_QUERY_DIM];
    char *user, *passwd, *url, *filename, *path, *name, *up_n, *down_n;
    const pstack_f *frame;
    int pol, sess, src_id, rid, downloaded;
    ftval val, ip;
    bool datafile;
    struct stat info;

    /* search pol and session */
    frame = ProtStackSearchProt(ppei->stack, pol_id);
    if (frame) {
        ProtGetAttr(frame, pol_polid_id, &val);
        pol = val.int32;
        ProtGetAttr(frame, pol_sesid_id, &val);
        sess = val.int32;
    }
    else {
        sess = pol = 1;
    }
    /* search source ip */
    src_id = -1;
    frame = ProtStackSearchProt(ppei->stack, ip_id);
    if (frame) {
        ProtGetAttr(frame, ip_src_id, &ip);
        src_id = DispHostSrch(&ip, FT_IPv4);
        if (src_id == -1) {
            /* search in db */
            FTString(&ip, FT_IPv4, flow_info);
            src_id = DispHostDb(flow_info, pol, sess);
            if (src_id == -1) {
                /* insert record */
                DnsDbSeach(&ip, FT_IPv4, query, XS_QUERY_DIM);
                src_id = DispHostDbIns(flow_info, query, pol, sess);
            }
            DispHostIns(&ip, FT_IPv4, src_id);
        }
    }
    else if (ipv6_id != -1) {
        frame = ProtStackSearchProt(ppei->stack, ipv6_id);
        if (frame) {
            ProtGetAttr(frame, ipv6_src_id, &ip);
            src_id = DispHostSrch(&ip, FT_IPv4);
            if (src_id == -1) {
                /* search in db */
                FTString(&ip, FT_IPv6, flow_info);
                src_id = DispHostDb(flow_info, pol, sess);
                if (src_id == -1) {
                    /* insert record */
                    DnsDbSeach(&ip, FT_IPv6, query, XS_QUERY_DIM);
                    src_id = DispHostDbIns(flow_info, query, pol, sess);
                }
                DispHostIns(&ip, FT_IPv6, src_id);
            }
        }
    }
    url = NULL;
    path = filename = passwd = user = "";
    cmpn = ppei->components;
    downloaded = TRUE;
    datafile = FALSE;
    while (cmpn != NULL) {
        if (cmpn->eid == pei_ftp_url_id) {
            url = cmpn->strbuf;
        }
        else if (cmpn->eid == pei_ftp_user_id) {
            user = cmpn->strbuf;
        }
        else if (cmpn->eid == pei_ftp_pswd_id) {
            passwd = cmpn->strbuf;
        }
        else if (cmpn->eid == pei_ftp_cmd_id) {
            filename = cmpn->name;
            path = cmpn->file_path;
        }
        else if (cmpn->eid == pei_ftp_file_in_id) {
            downloaded = 1;
            datafile = TRUE;
            filename = cmpn->name;
            path = cmpn->file_path;
        }
        else if (cmpn->eid == pei_ftp_file_out_id) {
            downloaded = 0;
            datafile = TRUE;
            filename = cmpn->name;
            path = cmpn->file_path;
        }
        else if (cmpn->eid == pei_ftp_file_offset_id) {
        }
        else if (cmpn->eid == pei_ftp_up_n_id) {
            up_n = cmpn->strbuf;
        }
        else if (cmpn->eid == pei_ftp_down_n_id) {
            down_n = cmpn->strbuf;
        }

        cmpn = cmpn->next;
    }
    
    /* compose query and insert record */
    if (url && path) {
        if (ppei->id == 0 && ppei->ret == TRUE) {
            /* new path */
            name = strrchr(path, '/');
            name++;
            sprintf(rep, XS_FTP_DIR_PATH"/%s", pol, sess, name);
            rename(path, rep);
            /* flow info */
            sprintf(flow_info, XS_FTP_DIR_PATH"/flow_%s.xml", pol, sess, name);
            DispFlowInfo(flow_info, ppei->stack);
            /* query */
            sprintf(query, XS_QUERY_FTP_TEMPLATE, sess, pol, src_id, time(NULL)-ppei->time_cap, flow_info, url,
                    user, passwd, rep);
            if (DispQuery(query, &rid) != 0) {
                printf("query: %s\n", query);
            }
            else {
                ppei->id = rid;
            }
        }
        if (ppei->ret == FALSE) {
            /* update number of file */
            sprintf(query, XS_QUERY_FTP_UPDATE, up_n, down_n, ppei->id);
            if (DispQuery(query, NULL) != 0) {
                printf("query: %s\n", query);
            }
        }
    }
    else if (datafile == TRUE && path && stat(path, &info) == 0) {
         /* new path */
        name = strrchr(path, '/');
        name++;
        sprintf(rep, XS_FTP_DIR_PATH"/%s", pol, sess, name);
        rename(path, rep);
        /* flow info */
        sprintf(flow_info, XS_FTP_DIR_PATH"/flow_%s.xml", pol, sess, name);
        DispFlowInfo(flow_info, ppei->stack);
        /* query */
        sprintf(query, XS_QUERY_FTP_DATA_TEMPLATE, sess, pol, src_id, time(NULL)-ppei->time_cap, flow_info,
                ppei->pid, filename, rep, info.st_size, downloaded);
        if (DispQuery(query, NULL) != 0) {
            printf("query: %s\n", query);
        }
    }

    return 0;
}


static int DispIpp(pei *ppei)
{
    pei_component *cmpn;
    char query[XS_QUERY_DIM];
    char pdf[XS_QUERY_DIM];
    char pcl[XS_QUERY_DIM];
    char flow_info[XS_STR_PATH];
    int pol, sess, src_id;
    char *url, *path, *name;
    off_t pdf_size, pcl_size;
    const pstack_f *frame;
    ftval val, ip;

    /* search pol and session */
    frame = ProtStackSearchProt(ppei->stack, pol_id);
    if (frame) {
        ProtGetAttr(frame, pol_polid_id, &val);
        pol = val.int32;
        ProtGetAttr(frame, pol_sesid_id, &val);
        sess = val.int32;
    }
    else {
        sess = pol = 1;
    }
    /* search source ip */
    src_id = -1;
    frame = ProtStackSearchProt(ppei->stack, ip_id);
    if (frame) {
        ProtGetAttr(frame, ip_src_id, &ip);
        src_id = DispHostSrch(&ip, FT_IPv4);
        if (src_id == -1) {
            /* search in db */
            FTString(&ip, FT_IPv4, flow_info);
            src_id = DispHostDb(flow_info, pol, sess);
            if (src_id == -1) {
                /* insert record */
                DnsDbSeach(&ip, FT_IPv4, query, XS_QUERY_DIM);
                src_id = DispHostDbIns(flow_info, query, pol, sess);
            }
            DispHostIns(&ip, FT_IPv4, src_id);
        }
    }
    else if (ipv6_id != -1) {
        frame = ProtStackSearchProt(ppei->stack, ipv6_id);
        if (frame) {
            ProtGetAttr(frame, ipv6_src_id, &ip);
            src_id = DispHostSrch(&ip, FT_IPv4);
            if (src_id == -1) {
                /* search in db */
                FTString(&ip, FT_IPv6, flow_info);
                src_id = DispHostDb(flow_info, pol, sess);
                if (src_id == -1) {
                    /* insert record */
                    DnsDbSeach(&ip, FT_IPv6, query, XS_QUERY_DIM);
                    src_id = DispHostDbIns(flow_info, query, pol, sess);
                }
                DispHostIns(&ip, FT_IPv6, src_id);
            }
        }
    }
    path = NULL;
    pdf[0] = '\0';
    pcl[0] = '\0';
    pdf_size = pcl_size = 0;
    cmpn = ppei->components;
    while (cmpn != NULL) {
        if (cmpn->eid == pei_ipp_url_id) {
            url = cmpn->strbuf;
        }
        if (cmpn->eid == pei_ipp_pdf_id) {
            path = cmpn->file_path;
            name = strrchr(path, '/');
            name++;
            sprintf(pdf, XS_IPP_DIR_PATH"/%s", pol, sess, name);
            rename(path, pdf);
            pdf_size = cmpn->file_size;
        }
        if (cmpn->eid == pei_ipp_pcl_id) {
            path = cmpn->file_path;
            name = strrchr(path, '/');
            name++;
            sprintf(pcl, XS_IPP_DIR_PATH"/%s", pol, sess, name);
            rename(path, pcl);
            pcl_size = cmpn->file_size;
        }
        cmpn = cmpn->next;
    }

    /* compose query and insert record */
    if (pcl[0] != '\0') {
        /* flow info */
        sprintf(flow_info, XS_IPP_DIR_PATH"/flow_%s.xml", pol, sess, name);
        DispFlowInfo(flow_info, ppei->stack);

        /* query */
        sprintf(query, XS_QUERY_PJL_TEMPLATE, sess, pol, src_id, time(NULL)-ppei->time_cap, flow_info,
                url, pdf, pdf_size, pcl, pcl_size);
        if (DispQuery(query, NULL) != 0) {
            printf("query: %s\n", query);
        }
    }
    
    return 0;
}


static int DispPjl(pei *ppei)
{
    pei_component *cmpn;
    char query[XS_QUERY_DIM];
    char pdf[XS_QUERY_DIM];
    char pcl[XS_QUERY_DIM];
    char flow_info[XS_STR_PATH];
    int pol, sess, src_id;
    char *url, *path, *name;
    off_t pdf_size, pcl_size;
    const pstack_f *frame;
    ftval val, ip;

    /* search pol and session */
    frame = ProtStackSearchProt(ppei->stack, pol_id);
    if (frame) {
        ProtGetAttr(frame, pol_polid_id, &val);
        pol = val.int32;
        ProtGetAttr(frame, pol_sesid_id, &val);
        sess = val.int32;
    }
    else {
        sess = pol = 1;
    }
    /* search source ip */
    src_id = -1;
    frame = ProtStackSearchProt(ppei->stack, ip_id);
    if (frame) {
        ProtGetAttr(frame, ip_src_id, &ip);
        src_id = DispHostSrch(&ip, FT_IPv4);
        if (src_id == -1) {
            /* search in db */
            FTString(&ip, FT_IPv4, flow_info);
            src_id = DispHostDb(flow_info, pol, sess);
            if (src_id == -1) {
                /* insert record */
                DnsDbSeach(&ip, FT_IPv4, query, XS_QUERY_DIM);
                src_id = DispHostDbIns(flow_info, query, pol, sess);
            }
            DispHostIns(&ip, FT_IPv4, src_id);
        }
    }
    else if (ipv6_id != -1) {
        frame = ProtStackSearchProt(ppei->stack, ipv6_id);
        if (frame) {
            ProtGetAttr(frame, ipv6_src_id, &ip);
            src_id = DispHostSrch(&ip, FT_IPv4);
            if (src_id == -1) {
                /* search in db */
                FTString(&ip, FT_IPv6, flow_info);
                src_id = DispHostDb(flow_info, pol, sess);
                if (src_id == -1) {
                    /* insert record */
                    DnsDbSeach(&ip, FT_IPv6, query, XS_QUERY_DIM);
                    src_id = DispHostDbIns(flow_info, query, pol, sess);
                }
                DispHostIns(&ip, FT_IPv6, src_id);
            }
        }
    }
    path = NULL;
    pdf[0] = '\0';
    pcl[0] = '\0';
    pdf_size = pcl_size = 0;
    cmpn = ppei->components;
    while (cmpn != NULL) {
        if (cmpn->eid == pei_pjl_url_id) {
            url = cmpn->strbuf;
        }
        if (cmpn->eid == pei_pjl_pdf_id) {
            path = cmpn->file_path;
            name = strrchr(path, '/');
            name++;
            sprintf(pdf, XS_PJL_DIR_PATH"/%s", pol, sess, name);
            rename(path, pdf);
            pdf_size = cmpn->file_size;
        }
        if (cmpn->eid == pei_pjl_pcl_id) {
            path = cmpn->file_path;
            name = strrchr(path, '/');
            name++;
            sprintf(pcl, XS_PJL_DIR_PATH"/%s", pol, sess, name);
            rename(path, pcl);
            pcl_size = cmpn->file_size;
        }
        cmpn = cmpn->next;
    }

    /* compose query and insert record */
    if (pcl[0] != '\0') {
        /* flow info */
        sprintf(flow_info, XS_PJL_DIR_PATH"/flow_%s.xml", pol, sess, name);
        DispFlowInfo(flow_info, ppei->stack);

        /* query */
        sprintf(query, XS_QUERY_PJL_TEMPLATE, sess, pol, src_id, time(NULL)-ppei->time_cap, flow_info,
                url, pdf, pdf_size, pcl, pcl_size);
        if (DispQuery(query, NULL) != 0) {
            printf("query: %s\n", query);
        }
    }
    
    return 0;
}


static int DispMms(pei *ppei)
{
    static int inc = 0;
    pei_component *cmpn;
    char query[XS_QUERY_DIM];
    char flow_info[XS_STR_PATH];
    char new_path[XS_QUERY_DIM];
    char *path, *name;
    const pstack_f *frame;
    char *from, *to, *cc, *bcc;
    int pol, sess, src_id, rid, contents;
    ftval val, ip;

    /* search pol and session */
    frame = ProtStackSearchProt(ppei->stack, pol_id);
    if (frame) {
        ProtGetAttr(frame, pol_polid_id, &val);
        pol = val.int32;
        ProtGetAttr(frame, pol_sesid_id, &val);
        sess = val.int32;
    }
    else {
        sess = pol = 1;
    }
    /* search source ip */
    src_id = -1;
    frame = ProtStackSearchProt(ppei->stack, ip_id);
    if (frame) {
        ProtGetAttr(frame, ip_src_id, &ip);
        src_id = DispHostSrch(&ip, FT_IPv4);
        if (src_id == -1) {
            /* search in db */
            FTString(&ip, FT_IPv4, flow_info);
            src_id = DispHostDb(flow_info, pol, sess);
            if (src_id == -1) {
                /* insert record */
                DnsDbSeach(&ip, FT_IPv4, query, XS_QUERY_DIM);
                src_id = DispHostDbIns(flow_info, query, pol, sess);
            }
            DispHostIns(&ip, FT_IPv4, src_id);
        }
    }
    else if (ipv6_id != -1) {
        frame = ProtStackSearchProt(ppei->stack, ipv6_id);
        if (frame) {
            ProtGetAttr(frame, ipv6_src_id, &ip);
            src_id = DispHostSrch(&ip, FT_IPv4);
            if (src_id == -1) {
                /* search in db */
                FTString(&ip, FT_IPv6, flow_info);
                src_id = DispHostDb(flow_info, pol, sess);
                if (src_id == -1) {
                    /* insert record */
                    DnsDbSeach(&ip, FT_IPv6, query, XS_QUERY_DIM);
                    src_id = DispHostDbIns(flow_info, query, pol, sess);
                }
                DispHostIns(&ip, FT_IPv6, src_id);
            }
        }
    }
    contents = 0;
    from = to = cc = bcc = " ";
    cmpn = ppei->components;
    while (cmpn != NULL) {
        if (cmpn->eid == pei_mms_from_id) {
            from = cmpn->strbuf;
        }
        else if (cmpn->eid == pei_mms_to_id) {
            to = cmpn->strbuf;
        }
        else if (cmpn->eid == pei_mms_cc_id) {
            cc = cmpn->strbuf;
        }
        else if (cmpn->eid == pei_mms_bcc_id) {
            bcc = cmpn->strbuf;
        }
        else if (cmpn->eid == pei_mms_part_id) {
            path = cmpn->file_path;
            name = strrchr(path, '/');
            name++;
            sprintf(new_path, XS_MMS_DIR_PATH"/%s", pol, sess, name);
            rename(path, new_path);
            DMemFree(cmpn->file_path);
            cmpn->file_path = DMemMalloc(strlen(new_path)+1);
            strcpy(cmpn->file_path, new_path);
            contents++;
        }
        else if (cmpn->eid == pei_mms_raw_id) {
            path = cmpn->file_path;
            name = strrchr(path, '/');
            name++;
            sprintf(new_path, XS_MMS_DIR_PATH"/%s", pol, sess, name);
            rename(path, new_path);
            DMemFree(cmpn->file_path);
            cmpn->file_path = DMemMalloc(strlen(new_path)+1);
            strcpy(cmpn->file_path, new_path);            
            contents++;
        }
        cmpn = cmpn->next;
    }
    /* flow info */
    sprintf(flow_info, XS_MMS_DIR_PATH"/flow_%lu_%i.xml", pol, sess, time(NULL), inc++);
    DispFlowInfo(flow_info, ppei->stack);
    /* compose query and insert record */
    sprintf(query, XS_QUERY_MMS_TEMPLATE, sess, pol, src_id, time(NULL)-ppei->time_cap, flow_info,
            " ", from, to, cc, bcc, contents);
    if (DispQuery(query, &rid) != 0) {
        printf("query: %s\n", query);
    }
    else {
        /* compose query and insert record for pevery art */
        cmpn = ppei->components;
        while (cmpn != NULL) {
            if (cmpn->eid == pei_mms_part_id) {
                if (cmpn->strbuf != NULL) {
                    if (cmpn->name != NULL) {
                        sprintf(query, XS_QUERY_MMS_CONT_TEMPLATE, sess, pol, src_id, rid, cmpn->strbuf, cmpn->name, cmpn->file_path, cmpn->file_size);
                    }
                    else {
                        sprintf(query, XS_QUERY_MMS_CONT_TEMPLATE, sess, pol, src_id, rid, cmpn->strbuf, "No name", cmpn->file_path, cmpn->file_size);
                    }
                }
                else if (cmpn->name != NULL) {
                    sprintf(query, XS_QUERY_MMS_CONT_TEMPLATE, sess, pol, src_id, rid, "unknow", cmpn->name, cmpn->file_path, cmpn->file_size);
                }
                else {
                    sprintf(query, XS_QUERY_MMS_CONT_TEMPLATE, sess, pol, src_id, rid, "unknow", "No name", cmpn->file_path, cmpn->file_size);
                }
                if (DispQuery(query, NULL) != 0) {
                    printf("query: %s\n", query);
                }
            }
            else if (cmpn->eid == pei_mms_raw_id) {
                sprintf(query, XS_QUERY_MMS_CONT_TEMPLATE, sess, pol, src_id, rid, "binary", "raw.mms", cmpn->file_path, cmpn->file_size);
                if (DispQuery(query, NULL) != 0) {
                    printf("query: %s\n", query);
                }
            }
            cmpn = cmpn->next;
        }
    }
    
    return 0;
}


static int DispTftp(pei *ppei)
{
    static int inc = 0;
    pei_component *cmpn;
    char query[XS_QUERY_DIM];
    char flow_info[XS_STR_PATH];
    char rep[XS_QUERY_DIM];
    char *url, *filename, *path, *name, *up_n, *down_n;
    const pstack_f *frame;
    int pol, sess, src_id, rid, downloaded;
    ftval val, ip;
    bool datafile;
    unsigned long data_size;

    /* search pol and session */
    frame = ProtStackSearchProt(ppei->stack, pol_id);
    if (frame) {
        ProtGetAttr(frame, pol_polid_id, &val);
        pol = val.int32;
        ProtGetAttr(frame, pol_sesid_id, &val);
        sess = val.int32;
    }
    else {
        sess = pol = 1;
    }
    /* search source ip */
    src_id = -1;
    frame = ProtStackSearchProt(ppei->stack, ip_id);
    if (frame) {
        ProtGetAttr(frame, ip_src_id, &ip);
        src_id = DispHostSrch(&ip, FT_IPv4);
        if (src_id == -1) {
            /* search in db */
            FTString(&ip, FT_IPv4, flow_info);
            src_id = DispHostDb(flow_info, pol, sess);
            if (src_id == -1) {
                /* insert record */
                DnsDbSeach(&ip, FT_IPv4, query, XS_QUERY_DIM);
                src_id = DispHostDbIns(flow_info, query, pol, sess);
            }
            DispHostIns(&ip, FT_IPv4, src_id);
        }
    }
    else if (ipv6_id != -1) {
        frame = ProtStackSearchProt(ppei->stack, ipv6_id);
        if (frame) {
            ProtGetAttr(frame, ipv6_src_id, &ip);
            src_id = DispHostSrch(&ip, FT_IPv4);
            if (src_id == -1) {
                /* search in db */
                FTString(&ip, FT_IPv6, flow_info);
                src_id = DispHostDb(flow_info, pol, sess);
                if (src_id == -1) {
                    /* insert record */
                    DnsDbSeach(&ip, FT_IPv6, query, XS_QUERY_DIM);
                    src_id = DispHostDbIns(flow_info, query, pol, sess);
                }
                DispHostIns(&ip, FT_IPv6, src_id);
            }
        }
    }
    
    url = NULL;
    path = filename = "";
    cmpn = ppei->components;
    downloaded = TRUE;
    datafile = FALSE;
    while (cmpn != NULL) {
        if (cmpn->eid == pei_tftp_url_id) {
            url = cmpn->strbuf;
        }
        else if (cmpn->eid == pei_tftp_cmd_id) {
            filename = cmpn->name;
            path = cmpn->file_path;
        }
        else if (cmpn->eid == pei_tftp_file_in_id) {
            downloaded = 1;
            datafile = TRUE;
            filename = cmpn->name;
            path = cmpn->file_path;
            data_size = cmpn->file_size;
        }
        else if (cmpn->eid == pei_tftp_file_out_id) {
            downloaded = 0;
            datafile = TRUE;
            filename = cmpn->name;
            path = cmpn->file_path;
            data_size = cmpn->file_size;
        }
        else if (cmpn->eid == pei_tftp_up_n_id) {
            up_n = cmpn->strbuf;
        }
        else if (cmpn->eid == pei_tftp_down_n_id) {
            down_n = cmpn->strbuf;
        }

        cmpn = cmpn->next;
    }
    
    /* compose query and insert record */
    if (url && path) {
        if (ppei->id == 0 && ppei->ret == TRUE) {
            /* new path */
            name = strrchr(path, '/');
            name++;
            sprintf(rep, XS_TFTP_DIR_PATH"/%s", pol, sess, name);
            DispCopy(path, rep, FALSE);
            /* flow info */
            sprintf(flow_info, XS_TFTP_DIR_PATH"/flow_%s.xml", pol, sess, name);
            DispFlowInfo(flow_info, ppei->stack);
            /* query */
            sprintf(query, XS_QUERY_TFTP_TEMPLATE, sess, pol, src_id, time(NULL)-ppei->time_cap, flow_info, url,
                    rep);
            if (DispQuery(query, &rid) != 0) {
                printf("query: %s\n", query);
            }
            else {
                ppei->id = rid;
            }
        }
        if (ppei->ret == FALSE) {
            /* update file */
            name = strrchr(path, '/');
            name++;
            sprintf(rep, XS_TFTP_DIR_PATH"/%s", pol, sess, name);
            DispCopy(path, rep, TRUE);
            /* update number of file */
            sprintf(query, XS_QUERY_TFTP_UPDATE, up_n, down_n, ppei->id);
            if (DispQuery(query, NULL) != 0) {
                printf("query: %s\n", query);
            }
        }
    }
    else if (datafile == TRUE && path) {
         /* new path */
        name = strrchr(path, '/');
        name++;
        sprintf(rep, XS_TFTP_DIR_PATH"/%s", pol, sess, name);
        DispCopy(path, rep, TRUE);
        /* flow info */
        sprintf(flow_info, XS_TFTP_DIR_PATH"/flow_%s.xml", pol, sess, name);
        DispFlowInfo(flow_info, ppei->stack);
        /* query */
        sprintf(query, XS_QUERY_TFTP_DATA_TEMPLATE, sess, pol, src_id, time(NULL)-ppei->time_cap, flow_info,
                ppei->pid, filename, rep, data_size, downloaded);
        if (DispQuery(query, NULL) != 0) {
            printf("query: %s\n", query);
        }
    }

    return 0;
}


static int DBInit(void)
{
    char *err;

    while ((db = sqlite_open(XS_DB_FILE, 0, &err)) == NULL) {
        sched_yield();
    }
    err = NULL;
    sqlite_exec(db, "PRAGMA synchronous=FULL", NULL, 0, 0);
#if 1
    if (sqlite_exec(db, "PRAGMA integrity_check", NULL, 0, &err) != 0) {
        if (err != NULL) {
            LogPrintf(LV_ERROR, "DB error: %s", err);
            free(err);
        }
    }
#endif
#if DB_ALWAYS_OPEN == 0
    sqlite_close(db);
    db = NULL;
#endif

    return 0;
}


static int DBClose(void)
{
#if DB_ALWAYS_OPEN
    sqlite_close(db);
#endif

    return 0;
}


int DispInit(void)
{
    pol_id = ProtId("pol");
    if (pol_id == -1) {
        printf("This dispacter run olny with pol dissector e capure\n");

        return -1;
    }
    pol_sesid_id = ProtAttrId(pol_id, "pol.sesid");
    pol_polid_id = ProtAttrId(pol_id, "pol.polid");
    pol_filename_id = ProtAttrId(pol_id, "pol.file");
    geo_id = 0;

    /* ip id */
    ip_id = ProtId("ip");
    if (ip_id != -1) {
        ip_dst_id = ProtAttrId(ip_id, "ip.dst");
        ip_src_id = ProtAttrId(ip_id, "ip.src");
    }
    ipv6_id = ProtId("ipv6");
    if (ipv6_id != -1) {
        ipv6_dst_id = ProtAttrId(ipv6_id, "ipv6.dst");
        ipv6_src_id = ProtAttrId(ipv6_id, "ipv6.src");
    }
 
    /* pei id */
    pop_id = ProtId("pop");
    if (pop_id != -1) {
        pei_pop_user_id = ProtPeiComptId(pop_id, "user");
        pei_pop_pswd_id = ProtPeiComptId(pop_id, "password");
        pei_pop_eml_id = ProtPeiComptId(pop_id, "eml");
    }

    smtp_id = ProtId("smtp");
    if (smtp_id != -1) {
        pei_smtp_to_id = ProtPeiComptId(smtp_id, "to");
        pei_smtp_from_id = ProtPeiComptId(smtp_id, "from");
        pei_smtp_eml_id = ProtPeiComptId(smtp_id, "eml");
    }

    imap_id = ProtId("imap");
    if (imap_id != -1) {
        pei_imap_user_id = ProtPeiComptId(imap_id, "user");
        pei_imap_pswd_id = ProtPeiComptId(imap_id, "password");
        pei_imap_eml_id = ProtPeiComptId(imap_id, "eml");
    }
    
    http_id = ProtId("http");
    if (http_id != -1) {
        pei_http_url_id = ProtPeiComptId(http_id, "url");
        pei_http_client_id = ProtPeiComptId(http_id, "client");
        pei_http_host_id = ProtPeiComptId(http_id, "host");
        pei_http_content_type_id = ProtPeiComptId(http_id, "content_type");
        pei_http_method_id = ProtPeiComptId(http_id, "method");
        pei_http_status_id = ProtPeiComptId(http_id, "status");
        pei_http_req_header_id = ProtPeiComptId(http_id, "req.header");
        pei_http_req_body_id = ProtPeiComptId(http_id, "req.body");
        pei_http_res_header_id = ProtPeiComptId(http_id, "res.header");
        pei_http_res_body_id = ProtPeiComptId(http_id, "res.body");
    }

    sip_id = ProtId("sip");
    if (sip_id != -1) {
        pei_sip_from_id = ProtPeiComptId(sip_id, "from");
        pei_sip_to_id = ProtPeiComptId(sip_id, "to");
    }

    ftp_id = ProtId("ftp");
    if (ftp_id != -1) {
        pei_ftp_url_id = ProtPeiComptId(ftp_id, "url");
        pei_ftp_user_id = ProtPeiComptId(ftp_id, "user");
        pei_ftp_pswd_id = ProtPeiComptId(ftp_id, "password");
        pei_ftp_cmd_id = ProtPeiComptId(ftp_id, "cmd");
        pei_ftp_file_in_id = ProtPeiComptId(ftp_id, "file_in");
        pei_ftp_file_out_id = ProtPeiComptId(ftp_id, "file_out");
        pei_ftp_file_offset_id = ProtPeiComptId(ftp_id, "offset");
        pei_ftp_down_n_id = ProtPeiComptId(ftp_id, "down_n");
        pei_ftp_up_n_id = ProtPeiComptId(ftp_id, "up_n");
    }

    ipp_id = ProtId("ipp");
    if (ipp_id != -1) {
        pei_ipp_url_id = ProtPeiComptId(ipp_id, "url");
        pei_ipp_pdf_id = ProtPeiComptId(ipp_id, "pdf");
        pei_ipp_pcl_id = ProtPeiComptId(ipp_id, "pcl");
    }

    pjl_id = ProtId("pjl");
    if (pjl_id != -1) {
        pei_pjl_url_id = ProtPeiComptId(pjl_id, "url");
        pei_pjl_pdf_id = ProtPeiComptId(pjl_id, "pdf");
        pei_pjl_pcl_id = ProtPeiComptId(pjl_id, "pcl");
    }

    mms_id = ProtId("mms");
    if (mms_id != -1) {
        pei_mms_url_id = ProtPeiComptId(mms_id, "url");
        pei_mms_from_id = ProtPeiComptId(mms_id, "from");
        pei_mms_to_id = ProtPeiComptId(mms_id, "to");
        pei_mms_cc_id = ProtPeiComptId(mms_id, "cc");
        pei_mms_bcc_id = ProtPeiComptId(mms_id, "bcc");
        pei_mms_part_id = ProtPeiComptId(mms_id, "part");
        pei_mms_raw_id = ProtPeiComptId(mms_id, "raw");
    }

    tftp_id = ProtId("tftp");
    if (tftp_id != -1) {
        pei_tftp_url_id = ProtPeiComptId(tftp_id, "url");
        pei_tftp_cmd_id = ProtPeiComptId(tftp_id, "cmd");
        pei_tftp_file_in_id = ProtPeiComptId(tftp_id, "file_in");
        pei_tftp_file_out_id = ProtPeiComptId(tftp_id, "file_out");
        pei_tftp_down_n_id = ProtPeiComptId(tftp_id, "down_n");
        pei_tftp_up_n_id = ProtPeiComptId(tftp_id, "up_n");
    }

    /* db connection */
    if (DBInit()) {
        printf("DB connections error\n");

        return -1;
    }
    pthread_mutex_init(&db_mux, NULL);
    pthread_mutex_init(&feed_mux, NULL);
    pthread_mutex_init(&host_mux, NULL);
    host_num = 0;
    host_dim = 0;
    host = NULL;
    
    return 0;
}


int DispEnd()
{
    if (geo_id != 0) {
        GearthClose(geo_id);
    }
    return DBClose();
}


int DispInsPei(pei *ppei)
{
    int ret;
    const pstack_f *frame;
    char kml_file[XS_STR_PATH];
    char *name;
    int pol, sess;
    ftval val;

    LogPrintf(LV_DEBUG, "Xplico System PEI dispatcher");

    if (ppei != NULL) {
        if (ppei->prot_id == pop_id) {
            ret = DispPop(ppei);
        }
        else if (ppei->prot_id == smtp_id) {
            ret = DispSmtp(ppei);
        }
        else if (ppei->prot_id == imap_id) {
            ret = DispImap(ppei);
        }
        else if (ppei->prot_id == http_id) {
            ret = DispHttp(ppei);
        }
        else if (ppei->prot_id == sip_id) {
            ret = DispSip(ppei);
        }
        else if (ppei->prot_id == ftp_id) {
            ret = DispFtp(ppei);
        }
        else if (ppei->prot_id == ipp_id) {
            ret = DispIpp(ppei);
        }
        else if (ppei->prot_id == pjl_id) {
            ret = DispPjl(ppei);
        }
        else if (ppei->prot_id == mms_id) {
            ret = DispMms(ppei);
        }
        else if (ppei->prot_id == tftp_id) {
            ret = DispTftp(ppei);
            PeiPrint(ppei);
        }
        else {
            PeiPrint(ppei);
        }
        if (geo_id == 0) {
            frame = ProtStackSearchProt(ppei->stack, pol_id);
            if (frame) {
                ProtGetAttr(frame, pol_polid_id, &val);
                pol = val.int32;
                ProtGetAttr(frame, pol_sesid_id, &val);
                sess = val.int32;
                ProtGetAttr(frame, pol_filename_id, &val);
                name = strrchr(val.str, '/');
                name++;
                sprintf(kml_file, XS_GEA_DIR_PATH"/%s.kml", pol, sess, name);
                FTFree(&val, FT_STRING);
                geo_id = 1;
                GearthNew(geo_id, kml_file);
                GearthPei(geo_id, ppei);
            }
        }
        else {
            GearthPei(geo_id, ppei);
        }
    }
    
    return 0;
}

