/* http.c
 * common functions to manage http packet
 *
 * $Id: http.c,v 1.1 2007/10/30 13:29:38 costa Exp $
 *
 * Xplico - Internet Traffic Decoder
 * By Gianluca Costa <g.costa@xplico.org>
 * Copyright 2007 Gianluca Costa & Andrea de Franceschi. Web: www.xplico.org
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "dmemory.h"
#include "http.h"
#include "log.h"

void HttpMsgPrint(const http_msg *msg)
{
    if (msg->uri) {
        LogPrintf(LV_DEBUG, "URI: %s", msg->uri);
    }
    if (msg->host) {
        LogPrintf(LV_DEBUG, "  Host: %s", msg->host);
    }
    if (msg->content_type) {
        LogPrintf(LV_DEBUG, "  Content-Type: %s", msg->content_type);
    }
    if (msg->content_encoding) {
        LogPrintf(LV_DEBUG, "  Content-Encoding: %s", msg->content_encoding);
    }
    if (msg->client) {
        LogPrintf(LV_DEBUG, "  Client: %s", msg->client);
    }
    if (msg->req_hdr_file) {
        LogPrintf(LV_DEBUG, "  Req Header: %s", msg->req_hdr_file);
    }
    if (msg->req_body_file) {
        LogPrintf(LV_DEBUG, "  Req Body: %s", msg->req_body_file);
    }
    if (msg->res_hdr_file) {
        LogPrintf(LV_DEBUG, "  Resp Header: %s", msg->res_hdr_file);
    }
    if (msg->res_body_file) {
        LogPrintf(LV_DEBUG, "  Resp Body: %s", msg->res_body_file);
    }
}


void HttpMsgFree(http_msg *msg)
{
    DMemFree(msg->uri);
    DMemFree(msg->host);
    DMemFree(msg->content_type);
    DMemFree(msg->content_encoding);
    DMemFree(msg->client);
    DMemFree(msg->req_hdr_file);
    DMemFree(msg->req_body_file);
    DMemFree(msg->res_hdr_file);
    DMemFree(msg->res_body_file);
}


void HttpMsgRemove(http_msg *msg)
{
    remove(msg->req_hdr_file);
    remove(msg->req_body_file);
    remove(msg->res_hdr_file);
    remove(msg->res_body_file);
}


void http_link(void)
{

}
