/* dbinterface.c
 *
 * $Id: dbinterface.c,v 1.2 2007/09/22 14:07:30 costa Exp $
 *
 * Xplico System
 * By Gianluca Costa <g.costa@xplico.org>
 * Copyright 2007 Gianluca Costa & Andrea de Franceschi. Web: www.xplico.org
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <mysql/mysql.h>
#include <sqlite.h> /* CakePHP is only compatible  with sqlite2 */

#include "dbinterface.h"


/* mysql */
static MYSQL *conn;                /* pointer to connection handler */
static sqlite *db_sqlite;          /* sqlite DB */
static DBType dbt;                 /* db type */

static int DBIntQuery(char *query)
{
    int ret;
    char *err = NULL;

    if (dbt == DB_MYSQL) {
        ret = mysql_query(conn, query);
    }
    else {
        while (sqlite_exec(db_sqlite, query, 0, 0, &err) == SQLITE_BUSY) {
            if (err != NULL)
                free(err);
            err = NULL;
        }
        ret = 0;
#warning "to be complete"
    }
    
    return ret;
}


int DBIntInit(DBType type)
{
    MYSQL *ret;
    char *err;

    dbt = type;

    if (dbt == DB_MYSQL) {
        /* mysql */
        conn = mysql_init (NULL);
        ret = mysql_real_connect (
            conn,          /* pointer to connection handler */
            DBINT_HOST_NAME,  /* host to connect to */
            DBINT_USER_NAME,  /* user name */
            DBINT_PASSWORD,   /* password */
            DBINT_DB_NAME,    /* database to use */
            0,             /* port (use default) */
            NULL,          /* socket (use default) */
            0);            /* flags (none) */
        
        if (!ret)
            return -1;
    }
    else if (dbt == DB_SQLITE) {
        /* sqlite */
        db_sqlite = sqlite_open(DBINT_DB_FILE, 0, &err);
        if (db_sqlite == 0) {
            return -1;
        }
    }
    else {
        return -1;
    }

    return 0;
}


int DBIntCapTime(int poi_id, int sol_id, cap_time *ctime)
{
    char query[DBINT_QUERY_DIM];

    if (dbt == DB_MYSQL) {
        /* query start time */
        sprintf(query, DBINT_1_QUERY_START_T_TEMPLATE, ctime->start_sec, sol_id);
        if (DBIntQuery(query) != 0) {
            return -1;
        }
        /* query end time */
        sprintf(query, DBINT_1_QUERY_END_T_TEMPLATE, ctime->end_sec, sol_id);
        if (DBIntQuery(query) != 0) {
            return -1;
        }
    }
    else {
        /* query start time */
        sprintf(query, DBINT_2_QUERY_START_T_TEMPLATE, time(NULL)-ctime->start_sec, sol_id);
        if (DBIntQuery(query) != 0) {
            return -1;
        }
        /* query end time */
        sprintf(query, DBINT_2_QUERY_END_T_TEMPLATE, time(NULL)-ctime->end_sec, sol_id);
        if (DBIntQuery(query) != 0) {
            return -1;
        }
    }

    return 0;
}


int DBIntDecStart(int poi_id, int sol_id)
{
    char query[DBINT_QUERY_DIM];

    /* query start decoding */
    sprintf(query, DBINT_QUERY_START_DEC, sol_id);
    if (DBIntQuery(query) != 0) {
        return -1;
    }

    return 0;
}


int DBIntDecEnd(int poi_id, int sol_id)
{
    char query[DBINT_QUERY_DIM];
    
    /* query end decoding */
    sprintf(query, DBINT_QUERY_END_DEC, sol_id);
    if (DBIntQuery(query) != 0) {
        return -1;
    }

    return 0;
}

int DBIntInputPcap(int poi_id, int sol_id, cap_time *ctime, unsigned long size,
                   const char *name, const char *md5, const char *sha1)
{
    char query[DBINT_QUERY_DIM];

    if (dbt == DB_MYSQL) {
        sprintf(query, DBINT_1_QUERY_INPUT_FILE, poi_id, sol_id, ctime->start_sec, ctime->end_sec,
                size, name, md5, sha1);
        if (DBIntQuery(query) != 0) {
            return -1;
        }
    }
    else {
        sprintf(query, DBINT_2_QUERY_INPUT_FILE, poi_id, sol_id,  time(NULL)-ctime->start_sec,
                time(NULL)-ctime->end_sec, size, name, md5, sha1);
        if (DBIntQuery(query) != 0) {
            return -1;
        }
    }

    return 0;
}

